from .core import ACGAN
from .core import BGAN
from .core import BIGAN
from .core import BaseGAN
from .core import CCGAN
from .core import CGAN
from .core import COGAN
from .core import DCGAN
from .core import DUALGAN
from .core import GAN
from .core import INFOGAN
from .core import LSGAN
from .core import SGAN
from .core import WGAN
